from __future__ import print_function, absolute_import

import os
import sys

import sh


def format_cwd(cwd):
    return os.path.join(cwd) if cwd else '.'


_original_init = sh.RunningCommand.__init__


def _verbose_init(self, cmd, call_args, stdin, stdout, stderr):
    encoding = sys.stdout.encoding

    print("[{cwd}]$ {command}".format(
        cwd=format_cwd(call_args['cwd']),
        command=(b' '.join(cmd[0].rsplit(b'/', 1)[1:] + cmd[1:])).decode(encoding),
    ))
    try:
        _original_init(self, cmd, call_args, stdin, stdout, stderr)
    except sh.ErrorReturnCode as e:
        sys.stdout.write(e.stdout.decode(encoding))
        sys.stderr.write(e.stderr.decode(encoding))
        raise
    else:
        sys.stdout.write(self.stdout.decode(encoding))
        sys.stderr.write(self.stderr.decode(encoding))


def patch_sh_verbose():
    sh.RunningCommand.__init__ = _verbose_init


class ShVerbose(object):
    def __init__(self, verbose=True):
        self.verbose = verbose
        self.start_init = None

    def __enter__(self):
        # record whatever the current __init__ is so we can reset it later
        self.start_init = sh.RunningCommand.__init__
        sh.RunningCommand.__init__ = (_verbose_init if self.verbose
                                      else _original_init)

    def __exit__(self, exc_type, exc_val, exc_tb):
        sh.RunningCommand.__init__ = self.start_init



from __future__ import print_function, absolute_import

from gevent import monkey

monkey.patch_all()

import argparse  # noqa E402
import gevent  # noqa E402
import jsonobject  # noqa E402
import os  # noqa E402
import re  # noqa E402
import sh  # noqa E402
import yaml  # noqa E402
from contextlib2 import ExitStack  # noqa E402

from .gitutils import (
    OriginalBranch,
    get_git,
    git_recent_tags,
    has_merge_conflict,
    print_merge_details,
)  # noqa E402

from .sh_verbose import ShVerbose  # noqa E402


class BranchConfig(jsonobject.JsonObject):
    trunk = jsonobject.StringProperty()
    name = jsonobject.StringProperty()
    branches = jsonobject.ListProperty(str)
    submodules = jsonobject.DictProperty(lambda: BranchConfig)
    pull_requests = jsonobject.ListProperty(str)

    def normalize(self):
        for submodule, subconfig in self.submodules.items():
            subconfig.trunk = subconfig.trunk or self.trunk
            subconfig.name = subconfig.name or self.name
            subconfig.normalize()

    def span_configs(self, path=('.',)):
        for submodule, subconfig in self.submodules.items():
            for item in subconfig.span_configs(path + (submodule,)):
                yield item
        yield os.path.join(*path), self

    def check_trunk_is_recent(self, path=None):
        # if it doesn't match our tag format
        if re.match(r'[\d-]+_[\d\.]+-\w+-deploy', self.trunk) is None:
            return True

        return self.trunk in git_recent_tags(path)


def fetch_remote(base_config, path, name="origin"):
    jobs = []
    seen = set()
    fetched = set()
    for path, config in base_config.span_configs((path,)):
        if path in seen:
            continue
        seen.add(path)
        git = get_git(path)
        remotes = set(git.remote().split())
        print("  [{cwd}] fetching {name}".format(cwd=path, name=name))
        jobs.append(gevent.spawn(git.fetch, name))
        for branch in (b for b in config.branches if ":" in b):
            remote, branch = branch.split(":", 1)
            if remote not in remotes:
                url = remote_url(git, remote)
                print("  [{path}] adding remote: {remote} -> {url}"
                      .format(**locals()))
                git.remote("add", remote, url)
                remotes.add(remote)
            print("  [{path}] fetching {remote} {branch}".format(**locals()))
            jobs.append(gevent.spawn(git.fetch, remote, branch))
            fetched.add(remote)

        for pr in config.pull_requests:
            print("  [{path}] fetching pull request {pr}".format(**locals()))
            pr = 'pull/{pr}/head:enterprise-{pr}'.format(pr=pr)
            jobs.append(gevent.spawn(git.fetch, 'origin', pr))

    gevent.joinall(jobs)
    print("fetched {}".format(", ".join(['origin'] + sorted(fetched))))


def remote_url(git, remote, original="origin"):
    origin_url = sh.grep(git.remote("-v"), original).split()[1]
    repo_name = origin_url.rsplit("/", 1)[1]
    return "https://github.com/{}/{}".format(remote, repo_name)


def has_ref(git, ref):
    """Return true if the named branch exists"""
    try:
        out = git("show-ref", "--verify", "--quiet", ref)
    except sh.ErrorReturnCode:
        return False
    return out.exit_code == 0


def has_local(git, branch):
    """Return true if the named local branch exists"""
    return has_ref(git, "refs/heads/{}".format(branch))


def has_remote(git, ref):
    """Return true if the named remote branch exists

    :param ref: Remote ref (example: origin/branch-name)
    """
    return has_ref(git, "refs/remotes/{}".format(ref))


def origin(branch):
    return "origin/{}".format(branch)


def sync_local_copies(config, path, push=True):
    base_config = config
    unpushed_branches = []

    def _count_commits(compare_spec):
        return int(sh.wc(git.log(compare_spec, '--oneline', _piped=True), '-l'))

    for path, config in base_config.span_configs((path,)):
        git = get_git(path)
        with OriginalBranch(git):
            for branch in [config.trunk] + config.branches:
                if ":" in branch or not has_local(git, branch):
                    continue
                git.checkout(branch)
                unpushed = _count_commits('origin/{0}..{0}'.format(branch))
                unpulled = _count_commits('{0}..origin/{0}'.format(branch))
                if unpulled or unpushed:
                    print(("  [{cwd}] {branch}: {unpushed} ahead "
                           "and {unpulled} behind origin").format(
                        cwd=path,
                        branch=branch,
                        unpushed=unpushed,
                        unpulled=unpulled,
                    ))
                else:
                    print("  [{cwd}] {branch}: Everything up-to-date.".format(
                        cwd=path,
                        branch=branch,
                    ))
                if unpushed:
                    unpushed_branches.append((path, branch))
                elif unpulled:
                    print("  Fastforwarding your branch to origin")
                    git.merge('--ff-only', origin(branch))
    if unpushed_branches and push:
        print("The following branches have commits that need to be pushed:")
        for path, branch in unpushed_branches:
            print("  [{cwd}] {branch}".format(cwd=path, branch=branch))
        exit(1)
    else:
        print("All branches up-to-date.")


def rebuild_staging(config, path, print_details=True, push=True):
    merge_conflicts = []
    not_found = []
    all_configs = list(config.span_configs((path,)))
    with ExitStack() as stack:
        for path, _ in all_configs:
            stack.enter_context(OriginalBranch(get_git(path)))
        for path, config in all_configs:
            git = get_git(path)
            try:
                git.checkout('-B', config.name, origin(config.trunk), '--no-track')
            except Exception:
                git.checkout('-B', config.name, config.trunk, '--no-track')
            for branch in config.branches:
                remote = ":" in branch
                if remote or not has_local(git, branch):
                    if remote:
                        remote_branch = branch.replace(":", "/", 1)
                    else:
                        remote_branch = origin(branch)
                    if not has_remote(git, remote_branch):
                        not_found.append((path, branch))
                        print("  [{cwd}] {branch} NOT FOUND".format(
                            cwd=format_cwd(path),
                            branch=branch,
                        ))
                        continue
                    branch = remote_branch
                print("  [{cwd}] Merging {branch} into {name}".format(
                    cwd=path,
                    branch=branch,
                    name=config.name
                ), end=' ')
                try:
                    git.merge(branch, '--no-edit')
                except sh.ErrorReturnCode_1:
                    merge_conflicts.append((path, branch, config))
                    try:
                        git.merge("--abort")
                    except sh.ErrorReturnCode_128:
                        pass
                    print("FAIL")
                else:
                    print("ok")
            for pr in config.pull_requests:
                branch = "enterprise-{pr}".format(pr=pr)
                print("  [{cwd}] Merging {pr} into {name}".format(
                    cwd=path,
                    pr=pr,
                    name=config.name
                ), end=' ')
                try:
                    git.merge(branch, '--no-edit')
                except sh.ErrorReturnCode_1:
                    merge_conflicts.append((path, branch, config))
                    try:
                        git.merge("--abort")
                    except sh.ErrorReturnCode_128:
                        pass
                    print("FAIL")
                else:
                    print("ok")
            if config.submodules:
                for submodule in config.submodules:
                    git.add(submodule)
                git.commit('-m', "update submodule refs", '--no-edit',
                           '--allow-empty')
        if push and not (merge_conflicts or not_found):
            for path, config in all_configs:
                # stupid safety check
                assert config.name != 'master', path
                print("  [{cwd}] Force pushing to origin {name}".format(
                    cwd=path,
                    name=config.name,
                ))
                force_push(get_git(path), config.name)

    if not_found:
        print("You must remove the following branches before rebuilding:")
        for cwd, branch in not_found:
            print("  [{cwd}] {branch}".format(
                cwd=format_cwd(cwd),
                branch=branch,
            ))
    if merge_conflicts:
        print("You must fix the following merge conflicts before rebuilding:")
        for cwd, branch, config in merge_conflicts:
            print("\n[{cwd}] {branch} => {name}".format(
                cwd=format_cwd(cwd),
                branch=branch,
                name=config.name,
            ))
            git = get_git(cwd)
            if print_details:
                print_conflicts(branch, config, git)

    if merge_conflicts or not_found:
        exit(1)


def print_conflicts(branch, config, git):
    if has_merge_conflict(branch, config.trunk, git):
        print(red("{} conflicts with {}".format(branch, config.trunk)))
        return

    conflict_found = False
    for other_branch in config.branches:
        if has_merge_conflict(branch, other_branch, git):
            print(red("{} conflicts with {}".format(branch, other_branch)))
            conflict_found = True

    if not conflict_found:
        print_merge_details(branch, config.name, git,
                            known_branches=config.branches)


def force_push(git, branch):
    try:
        git.push('origin', branch, '--force')
    except sh.ErrorReturnCode_128 as e:
        # oops we're using a read-only URL, so change to the suggested url
        try:
            line = sh.grep(git.remote("-v"),
                           '-E', r'^origin.(https|git)://github\.com/.*\(push\)$')
        except sh.ErrorReturnCode_1:
            raise e
        old_url = line.strip().split()[1]
        prefix = "git" if old_url.startswith("git:") else "https"
        new_url = old_url.replace(prefix + "://github.com/", "git@github.com:")
        print("    {} -> {}".format(old_url, new_url))
        git.remote('set-url', 'origin', new_url)
        git.push('origin', branch, '--force')


def format_cwd(cwd):
    return os.path.join(cwd) if cwd else '.'


class DisableGitHooks(object):
    already_disabled = None

    def __init__(self, path='.git/hooks'):
        import uuid
        self.path = path
        self.guid = uuid.uuid4().hex

    @property
    def hidden_path(self):
        return self.path + '-' + self.guid

    def __enter__(self):
        try:
            sh.test('-d', self.path)
            self.already_disabled = False
        except sh.ErrorReturnCode_1:
            self.already_disabled = True
        else:
            sh.mv(self.path, self.hidden_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.already_disabled:
            sh.mv(self.hidden_path, self.path)


def _wrap_with(code):

    def inner(text, bold=False):
        c = code

        if bold:
            c = "1;%s" % c
        return "\033[%sm%s\033[0m" % (c, text)
    return inner


red = _wrap_with('31')


def main():
    parser = argparse.ArgumentParser(description='Rebuild the deploy branch for an environment')
    parser.add_argument("config_path", help="Path to the YAML configuration file")
    parser.add_argument("actions", nargs="*")
    parser.add_argument("-p", "--path", default=".", help="Path to the repository")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("--push", action="store_true", help="Push the changes to remote git repository.")
    args = parser.parse_args()

    git = get_git()
    print("Fetching master")
    git.fetch("origin", "master")
    if args.push:
        print("Checking branch config for modifications")
        if git.diff("origin/master", "--", args.config_path):
            print(red("'{}' on this branch different from the one on master".format(args.config_path)))
            exit(1)

    with open(args.config_path) as config_yaml:
        config = yaml.safe_load(config_yaml)

    code_root = os.path.abspath(args.path)
    if not os.path.exists(code_root):
        print(red("Repository path does not exist: {}".format(code_root)))
        exit(1)

    repo_config = BranchConfig.wrap(config)
    repo_config.normalize()

    if not repo_config.check_trunk_is_recent(code_root):
        print("The trunk is not based on a very recent commit")
        print("Consider using one of the following:")
        print(git_recent_tags(code_root))
        exit(1)

    if not args.actions:
        args.actions = 'fetch sync rebuild'.split()
    with DisableGitHooks(), ShVerbose(args.verbose):
        print("\nRebuilding '{}' branch.".format(repo_config.name))
        if 'fetch' in args.actions:
            fetch_remote(repo_config, code_root)
        if 'sync' in args.actions:
            sync_local_copies(repo_config, code_root, push=args.push)
        if 'rebuild' in args.actions:
            rebuild_staging(repo_config, code_root, push=args.push)


if __name__ == "__main__":
    main()



from __future__ import print_function, absolute_import

import argparse
import os

import sh
from sh import ErrorReturnCode

from .checkyaml import checkyaml, YamlError
from .gitutils import get_git, get_grep
from .sh_verbose import ShVerbose

grep = get_grep()
git = get_git()


def get_branch():
    branch = sh.sed(grep(git.branch(), "^\\*"), "s/* //")
    return branch.stdout.strip().decode()


def main():
    parser = argparse.ArgumentParser(description="Safely commit a single file to git.")
    parser.add_argument("files", nargs="+", help="Path to the files")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("--push", action="store_true", help="Push the changes to remote git repository.")
    args = parser.parse_args()

    files = args.files

    with ShVerbose(args.verbose):

        branch = get_branch()
        if branch != "master":
            print("You may only commit a deploy branch config file to master. '{}'".format(branch))
            exit(1)

        git.add(*files)
        try:
            staged = sh.grep(git.diff("--staged", "--stat"), "|")
        except ErrorReturnCode:
            print("You have no changes to commit.")
            exit(1)

        staged_files = filter(None, [line.split("|")[0].strip() for line in staged.split("\n")])
        if not staged_files:
            print("You have no changes to commit.")
            exit(1)

        basenames = {os.path.basename(filename) for filename in files}
        staged_basenames = {os.path.basename(filename) for filename in staged_files}
        if basenames != staged_basenames:
            print("Unexpected files staged: {}".format(", ".join(staged_files)))
            exit(1)

        for filename in files:
            if os.path.splitext(filename)[1].lower() in ('yaml', 'yml'):
                try:
                    checkyaml(filename)
                except YamlError as e:
                    print("Yaml error in file:")
                    print(e)
                    exit(1)

        git.fetch()
        if git.log("--max-count=1", "origin/{0}..{0}".format(branch)).strip():
            print("Your local '{0}' is ahead of 'origin/{0}'.".format(branch))

        message = "updating files: {}".format(", ".join(files))
        git.commit("--message", message, "--message", "[ci skip]")

        if args.push:
            git.push("origin", branch)


if __name__ == "__main__":
    main()



from __future__ import print_function, absolute_import

import io
import yaml


class YamlError(Exception):
    pass


def checkyaml(filename):
    try:
        yaml.safe_load(io.open(filename, encoding='utf-8'))
    except yaml.YAMLError as e:
        message = "Error in file {} ".format(filename)
        if hasattr(e, "problem_mark"):
            mark = e.problem_mark
            message += "on line {} (column {}):".format(mark.line + 1, mark.column + 1)
            f = io.open(filename, encoding='utf-8')
            for _ in range(mark.line + 1):
                message += "\n" + "    " + f.readline()
            message += "    " + (" " * mark.column) + "^"
        raise YamlError(message)



import re

import sh

from .sh_verbose import ShVerbose


def get_git(path=None):
    return sh.git.bake(_tty_out=False, _cwd=path)


def get_grep():
    return sh.grep.bake(_tty_out=False)


def get_tail():
    return sh.tail.bake(_tty_out=False)


class OriginalBranch(object):
    def __init__(self, git=None):
        self.git = git or get_git()
        self.original_branch = None

    def __enter__(self):
        self.original_branch = git_current_branch(self.git)
        return self.original_branch

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.git.checkout(self.original_branch)
        except Exception as err:
            print("cannot checkout '{}': {}".format(self.original_branch, err))


def git_current_branch(git=None):
    git = git or get_git()
    grep = get_grep()
    branch = grep(git.branch(), '^* ').strip()[2:]
    if branch.startswith('('):
        branch = git.log('--pretty=oneline', n=1).strip().split(' ')[0]
    return branch


def git_recent_tags(grep_string="production-deploy", path=None):
    git, grep, tail = get_git(path), get_grep(), get_tail()
    last_tags = tail(grep(git.tag('--sort=committerdate'), grep_string), n=4)
    return last_tags


def git_submodules(git=None):
    git = git or get_git()
    submodules = []
    for line in git.submodule().split('\n')[:-1]:
        path = line[1:].split()[1]
        submodules.append(path)
    return submodules


def git_check_merge(branch1, branch2, git=None):
    """
    returns True if branch1 would auto-merge cleanly into branch2,
    False if the merge requires human assistance

    Thanks to http://stackoverflow.com/a/501461/240553

    """
    git = git or get_git()
    with ShVerbose(False):
        orig_branch = git_current_branch(git)
        git.checkout(branch2)
        is_behind = git.log('{0}..{1}'.format(branch2, branch1),
                            max_count=1).strip()
        clean_merge = True
        if is_behind:
            try:
                git.merge('--no-commit', '--no-ff', branch1).strip()
            except sh.ErrorReturnCode_1:
                # git merge returns 1 when there's a conflict
                clean_merge = False
            git.merge('--abort')
        git.checkout(orig_branch)
        return clean_merge


def has_merge_conflict(branch1, branch2, git):
    return not git_check_merge(branch1, branch2, git=git)


def git_bisect_merge_conflict(branch1, branch2, git=None):
    """
    return the branch2 commit that prevents branch1 from being merged in

    """
    git = git or get_git()
    grep = get_grep()
    with OriginalBranch(git):
        try:
            base = git('merge-base', branch1, branch2).strip()
            if git_check_merge(branch1, branch2, git):
                return None
            assert git_check_merge(branch1, base, git)
            git.bisect('reset')
            txt = git.bisect('start', branch2, base, '--')
            while 'is the first bad commit' not in txt:
                commit = git_current_branch(git)
                if git_check_merge(branch1, commit, git):
                    txt = git.bisect('good')
                else:
                    txt = git.bisect('bad')
            try:
                # txt has a line that's like "<commit> is the first bad commit"
                return grep(txt, ' is the first bad commit$').strip().split(' ')[0]
            except sh.ErrorReturnCode_1:
                raise Exception('Error finding offending commit: '
                                '"^commit" does not match\n{}'.format(txt))
        finally:
            git.bisect('reset')


def _left_pad(padding, text):
    return padding + ('\n' + padding).join(text.split('\n'))


def print_one_way_merge_details(branch1, branch2, git, known_branches=None):
    def format_branch(remote, branch):
        return branch if remote == 'origin' else '{}/{}'.format(remote, branch)

    if known_branches is None:
        # make `foo in known_branches` always return True
        class InfiniteSet(object):
            def __contains__(self, item):
                return True
        known_branches = InfiniteSet()

    commit = git_bisect_merge_conflict(branch1, branch2, git)
    if commit:
        print('  * First conflicting commit on {0}:\n'.format(branch2))
        print(_left_pad(' ' * 4, git.log('-n1', commit)))
        branches = git.branch('--remote', '--contains', commit)
        other_branches = [
            format_branch(*b)
            for b in re.findall(r'([a-zA-Z0-9-]*)/([\w+-]*)', str(branches))
            if b[0] != 'origin' or (b[1] != branch2 and b[1] in known_branches
                                    and b[1] != 'HEAD')
        ]
        if other_branches:
            msg = 'This commit also appears on these branches:'
            print(_left_pad(' ' * 4, msg))
            for branch in other_branches:
                print(_left_pad(' ' * 4, '* {}'.format(branch)))
    else:
        print('  * No conflicting commits on {0}'.format(branch2))


def print_merge_details(branch1, branch2, git, known_branches=None):
    print_one_way_merge_details(branch1, branch2, git,
                                known_branches=known_branches)
    print_one_way_merge_details(branch2, branch1, git,
                                known_branches=known_branches)


if __name__ == '__main__':
    import sys
    args = sys.argv[1:]
    options = ['show-conflict']
    try:
        option = args.pop(0)
    except IndexError:
        option = None
    if option == 'show-conflict':
        if len(args) == 2:
            print_merge_details(*args, git=get_git())
        else:
            print ('usage: python scripts/gitutils.py '
                   'show-conflict <branch1> <branch2>')
    else:
        print('usage: python scripts/gitutils.py <command> [args...]\n')
        print('Available commands:')
        print(_left_pad('   ', '\n'.join(options)))



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'Click>=7.0',
    'gevent>=1.4.0',
    'jsonobject>=0.9.9',
    'sh>=1.0.9',
    'PyYAML>=5.1',
    'contextlib2>=0.5.5',
]

setup_requirements = ['pytest-runner', ]

setup(
    author="Dimagi",
    author_email='dev@dimagi.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Utility tool for building Git branches my merging multiple other branches together.",
    entry_points={
        'console_scripts': [
            'git-build-branch=git_build_branch.branch_builder:main',
            'safe-commit-files=git_build_branch.safe_commit_files:main',
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme,
    include_package_data=True,
    keywords='git-build-branch',
    name='git-build-branch',
    packages=find_packages(include=['git_build_branch', 'git_build_branch.*']),
    setup_requires=setup_requirements,
    url='https://github.com/dimagi/git-build-branch',
    version='0.1.10',
    zip_safe=False,
)
